/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  flatten.P
  Author(s)		:  Saumya Debray, Terrance Swift, Kostis Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  May 23, 1993
========================================================================*/


/*======================================================================*/
/* flatten(+ModuleIn, -ModuleOut)					*/
/*	Predicate flatten/2 will flatten a clause so that it contains	*/
/*	no compound term with an argument that is again compound term.	*/
/*	We say that this kind of clauses are 1-level functor allowance.	*/
/*	However, the procedure here is more general, in that you can	*/
/*	transform clauses to k-level functor allowance form, where	*/
/*	k >= 0 (when k = 0, 1-level function is still allowed in '='/2).*/
/*	The only thing you need to change is to redefine the predicate	*/
/*	level_allowed/1.						*/
/*	Currently, the arguments of is/2 are not flattened due to 	*/
/*	the special processing of the compiler. If some day I implement	*/
/*	functions, is/2 can then be replaced by =/2, the exception will	*/
/*	be no longer needed.						*/
/*	For type inference procedure, I test with level_allowed(1);	*/
/*	but for compiling, it REQUIRES level_allowed(1).		*/
/*======================================================================*/

level_allowed(1).

flatten(module(SymTab,TyList,DcList,FnList,ClList,QrList),
	module(SymTab,TyList,DcList,FnList,ClList1,QrList1)) :-
	flatten0(ClList, ClList1, SymTab),
	flatten0(QrList, QrList1, SymTab).

/* The parameter SymTab is used only for function translation */

flatten0([], [], _SymTab).
flatten0([One|CList0], [One1|CList1], SymTab) :-
	flt_block(One, One1, SymTab),
	flatten0(CList0, CList1, SymTab).

flt_block(pred(Sym,Pcs0,_), pred(Sym,Pcs1,_), SymTab) :-
	flt_clauses(Pcs0, Pcs1, SymTab).
% flt_block(fun(Sym, CList0, _), fun(Sym, CList1, _), SymTab) :-
% 	flt_clauses(CList0, CList1, SymTab).

flt_clauses([], [], _SymTab).
flt_clauses([clause(Args,B0,P)|Pcs0], [clause(NArgs,B1,P)|Pcs1], SymTab) :-
	flt_args(Args, NArgs, B2, B1, 1, 0, SymTab),
	flt_body(B0, B2, SymTab),
/*	flt_writeclause(clause(NArgs,B1,P)), */
	flt_clauses(Pcs0, Pcs1, SymTab).

flt_body(and(G1,G2), and(G3,G4), SymTab) :- 
	flt_body(G1, G3, SymTab), flt_body(G2, G4, SymTab).
flt_body(or(G1,G2), or(G3,G4), SymTab) :-
	flt_body(G1, G3, SymTab), flt_body(G2, G4, SymTab).
flt_body(if(G1,G2), if(G3,G4), SymTab) :- 
	flt_body(G1, G3, SymTab), flt_body(G2, G4, SymTab).
flt_body(not(G1), not(G2), SymTab) :- flt_body(G1, G2, SymTab).
			% The body is guarenteed not be a var, by parser
flt_body('_$complete'(G1),'_$complete'(G2),SymTab) :- flt_body(G1, G2, SymTab).
			% The body is guarenteed not be a var, by parser
flt_body(goal(Sym, Args), NewGoal, SymTab) :- 
	flt_args(Args, NArgs, goal(Sym, NArgs), NewGoal, 1, 0, SymTab).
flt_body(inlinegoal(P, A, Args), NewGoal, SymTab) :- 
	( P = 'is', A = 2 -> InIs = 1 ; InIs = 0 ),
	flt_args(Args, NArgs, inlinegoal(P,A,NArgs), NewGoal, 1, InIs, SymTab).
% flt_body(funbody(Res, Constr), funbody(Res1, Constr1), SymTab) :-
% 	Constr1 = and(Constr2, Constr3),
% 	flt_arg(Res, Res1, inlinegoal('_$return',1,[Res1]), Constr3,1,0,SymTab),
% 	flt_body(Constr, Constr2, SymTab).

/*======================================================================*/
/* flt_arg(+OldArg, -NewArg, +OldGoals, -NewGoals, +Level, +InIs)	*/
/*									*/
/*	InIs| Level>allowd   Level<=allowd				*/
/*	-----------------------------------				*/
/*	0   |  no flatten   flatten to '='				*/
/*	1   |  no flatten   flatten to 'is'				*/
/*									*/
/*	When function translation is allowed, all functions are		*/
/*	flattened except in the case when a function occurs within	*/
/*	"is/2" at the allowed level.					*/
/*======================================================================*/

flt_arg(varocc(Var), varocc(Var), OldGoal, OldGoal, _, _InIs, _SymTab).
flt_arg(integer(Int), integer(Int), OldGoal, OldGoal, _, _InIs, _SymTab).
flt_arg(real(V), real(V), OldGoal, OldGoal, _, _InIs, _SymTab).
flt_arg(string(S), string(S), OldGoal, OldGoal, _, _InIs, _SymTab).
flt_arg(constant(Sym), constant(Sym), OldGoal, OldGoal, _, _InIs, _SymTab).
flt_arg(structure(Sym,Args), New,OldGoal,NewGoal,Level,InIs, SymTab) :-
	level_allowed(Allowed),
	NStr = structure(Sym, NArgs),
	Level1 is Level + 1,
	flt_args(Args, NArgs, MedGoal, NewGoal, Level1, InIs, SymTab),
	( sym_prop((fun), Sym) ->
	    sym_name(Sym, F, A),
	    ( InIs =:= 1, Level =< Allowed ->
			New = NStr, MedGoal = OldGoal
	    ; gennum(Vid), New = varocc(Vid),
		( builtin_function(F, A, _) ->
		      MedGoal = and(inlinegoal(is, 2, [New,NStr]), OldGoal)
		;     MedGoal = and(fungoal(Sym, New, NArgs), OldGoal)
		)
	    )
	; ( Level =< Allowed, New = NStr, MedGoal = OldGoal
	  ; Level > Allowed, gennum(Vid), New = varocc(Vid),
	    ( InIs =:= 1,
		MedGoal = and(inlinegoal(is, 2, [New,NStr]), OldGoal)
	    ; InIs =\= 1, 
		MedGoal = and(inlinegoal('=', 2, [New,NStr]), OldGoal)
	    )
	  )
	).

flt_args([], [], OldGoal, OldGoal, _, _InIs, _SymTab).
flt_args([Arg|Args], [NArg|NArgs], OldGoal, NewGoal, Level, InIs, SymTab) :- 
	flt_arg(Arg, NArg, OldGoal, MedGoal, Level, InIs, SymTab),
	flt_args(Args, NArgs, MedGoal, NewGoal, Level, InIs, SymTab).

/*=== used for debugging ===================================*/

/*

:- import ttywrite/1, ttywritenl/1 from standard.

flt_writeclause(clause(Args,funbody(Body,Constr),Prag)) :-
	ttywrite('<fun>('),
	flt_writeargs(Args),
	ttywrite(') = '),
	flt_writearg(Body),
	ttywrite(' :- '),
	flt_writegoal(Constr),
	ttywritenl('. ').
flt_writeclause(clause(Args,Body,Prag)) :-
	ttywrite('<pred>('),
	flt_writeargs(Args),
	ttywrite(') :- '),
	flt_writegoal(Body),
	ttywritenl('. ').
	
flt_writegoal(and(X, Y)) :-
	flt_writegoal(X),
	ttywrite(', '),
	flt_writegoal(Y).
flt_writegoal(goal(Sym, Args)) :-
	sym_name(Sym, Name, _),
	ttywrite(Name),
	ttywrite('('),
	flt_writeargs(Args),
	ttywrite(')').
flt_writegoal(inlinegoal(Name, _, Args)) :-
	ttywrite(Name),
	ttywrite('('),
	flt_writeargs(Args),
	ttywrite(')').
flt_writegoal(fungoal(Sym, Res, Args)) :-
	flt_writearg(Res),
	ttywrite(' := '),
	sym_name(Sym, Name, _),
	ttywrite(Name),
	ttywrite('('),
	flt_writeargs(Args),
	ttywrite(')').

flt_writeargs([]).
flt_writeargs([One]) :-
	flt_writearg(One).
flt_writeargs([One|Rest]) :-
	flt_writearg(One),
	ttywrite(', '),
	flt_writeargs(Rest).

flt_writearg(varocc(ID)) :- ttywrite('_'), ttywrite(ID).
flt_writearg(constant(Sym)) :- sym_name(Sym,Name,_), ttywrite(Name).
flt_writearg(integer(I)) :- ttywrite(I).
flt_writearg(structure(Sym, Args)) :-
	sym_name(Sym, Name, _),
	ttywrite(Name),
	ttywrite('('),
	flt_writeargs(Args),
	ttywrite(')').
*/
