/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tp_eval.P
  Author(s)		:  Saumya Debray, Kostis Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  August 1, 1993
======================================================================*/


/*======================================================================*/
/* eval_exp(+Term, -R, -Inst, -InstTail, #HoldRegs, +Tin, Tout)		*/
/*	R is the register allocated for the result			*/
/* Generate code for an evaluable term.					*/
/* The code generated for different terms are listed below.		*/
/*  integer(N):   putnumcon(R, N)			Res in R	*/
/*  real(N):	  putfloate(R, N)			in R		*/
/*  constant(S):  error							*/
/*  varocc(_,_):  movereg(Loc,R)/none			in Loc/R	*/
/*  structure cases ......						*/
/*     + X:	  code(X, R)				in R		*/
/*     - X:	  putnumcon(R,0),code(X,R1),subreg(R1,R) in R		*/
/*     \ X:	  code(X, R), negate(R)			in R		*/
/*   other unary builtin functions:					*/
/*     f(X):	  code(X, R), unifunc(FunNum,R)		in R		*/
/*     X mod Y:	  code(X,R0),code(Y,R1),movreg(R0,R),idivreg(R1,R0)	*/
/*		  mulreg(R1,R0),subreg(R0,R) 		in R		*/
/*   other bindary builtin functions:					*/
/*     f(X,Y):	  code(X,R1),code(Y,R2),Corres. Inst.	in R1		*/
/*     ... ... 								*/
/*======================================================================*/

eval_exp(varocc(_,Prag),R,Pil,Pilr,_,Tin,Tout) :-
	eval_var(Prag,R,Pil,Pilr,Tin,Tout).
eval_exp(integer(N),R,[Inst|Pil],Pil,_,Tin,Tout) :- 
	getreg(Tin,R),
	hold(R,Tin,Tout),
	Inst = putnumcon(R,N).
eval_exp(real(N),R,[Inst|Pil],Pil,_,Tin,Tout) :- 
	getreg(Tin,R),
	hold(R,Tin,Tout),
	Inst = putfloat(R,N).
eval_exp(structure(Sym,Args),R,Pil,Pilr,HoldR,Tin,Tout) :-
	sym_name(Sym, Op, Arity),
	( Arity =:= 2 -> Args = [E1,E2],
		eval_binop(Op,E1,E2,R,Pil,Pilr,HoldR,Tin,Tout)
	; Arity =:= 1 -> Args = [E],
		eval_uniop(Op, E,R,Pil,Pilr,HoldR,Tin,Tout)
	; error(('Unknown function: ', Op, ' in is/2'))
	).
eval_exp(constant(_S),R,[Inst|Pil],Pil,_,Tin,Tout) :-	/* may never occur */
	error(('non-numeric constant in numeric expression: ',_N)),
	Inst = fail, Tin = Tout, R = 0.

eval_uniop('+',E,R,Pil,Pilr,HoldR,Tin,Tout) :-	/* unary plus */
    eval_exp(E,R,Pil,Pilr,HoldR,Tin,Tout).
eval_uniop('-',E,R,Pil,Pilr,HoldR,Tin,Tout) :-	/* unary minus */
    getreg(Tin,R), hold(R,Tin,Tmid1),
    Pil = [putnumcon(R,0)|Pilm],
    eval_exp(E,R1,Pilm,[subreg(R1,R)|Pilr],HoldR,Tmid1,Tmid2),
    release(R,Tmid2,Tmid3),
    release_if_done0(E,R1,HoldR,Tmid3,Tout).
eval_uniop('\',E,R,Pil,Pilr,HoldR,Tin,Tout) :-  /* negation */
    eval_exp(E,R,Pil,[negate(R)|Pilr],HoldR,Tin,Tout).
eval_uniop(Op,E,R,Pil,Pilr,HoldR,Tin,Tout) :-   /* others */
    builtin_function(Op, 1, Num),
    eval_exp(E,R,Pil,[unifunc(Num,R)|Pilr],HoldR,Tin,Tout).
eval_uniop(Op,_,_,Pil,Pil,_,Tin,Tin) :-
    error(('Unknown function: ', Op, '/1 in is/2')).

eval_binoptab(+,R1,R2,addreg(R2,R1)).
eval_binoptab(-,R1,R2,subreg(R2,R1)).
eval_binoptab(*,R1,R2,mulreg(R2,R1)).
eval_binoptab(/,R1,R2,divreg(R2,R1)).
eval_binoptab(//,R1,R2,idivreg(R2,R1)).
eval_binoptab('/\',R1,R2,and(R2,R1)).
eval_binoptab('\/',R1,R2,or(R2,R1)).
eval_binoptab('<<',R1,R2,logshiftl(R2,R1)).
eval_binoptab('>>',R1,R2,logshiftr(R2,R1)).

eval_binop('+').
eval_binop('-').
eval_binop('*').
eval_binop('/').
eval_binop('//').
eval_binop('/\').
eval_binop('\/').
eval_binop('<<').
eval_binop('>>').

arithrelop('=:=').
arithrelop('=\=').
arithrelop('<').
arithrelop('=<').
arithrelop('>').
arithrelop('>=').

eval_arithreloptab(=:=,R,Label,jumpnz(R,Label)).
eval_arithreloptab(=\=,R,Label,jumpz(R,Label)).
eval_arithreloptab(<,R,Label,jumpge(R,Label)).
eval_arithreloptab(=<,R,Label,jumpgt(R,Label)).
eval_arithreloptab(>,R,Label,jumple(R,Label)).
eval_arithreloptab(>=,R,Label,jumplt(R,Label)).

eval_var(Prag,R,Pil,Pilr,Tin,Tout) :-
    alloc_reg(Prag,Tin,Tmid),
    vprag_occ(Prag,Occ),
    ( (Occ=s; Occ=l), eval_numgetinst(Prag,R,Pil,Pilr,Tmid,Tout)
    ; error('unbound variable in arithmetic expression'),
      Pil = [fail|Pilr], R = 0, Tmid = Tout
    ).

eval_binop(Op,E1,E2,R1,Pil,Pilr,HoldR,Tin,Tout) :-
    eval_binop(Op),
    eval_exp(E1,R1,Pil,Pilm,HoldR,Tin,Tmid1),
    eval_exp(E2,R2,Pilm,[Inst|Pilr],HoldR,Tmid1,Tmid2),
    eval_binoptab(Op,R1,R2,Inst),
    release_if_done0(E2,R2,HoldR,Tmid2,Tout).
eval_binop(mod,E1,E2,R,Pil,Pilr,HoldR,Tin,Tout) :-
    eval_exp(E1,R0,Pil,Pilm0,HoldR,Tin,Tmid),
    eval_exp(E2,R1,Pilm0,Pilm1,HoldR,Tmid,Tmid1),
    getreg(Tmid1,R),
    Pilm1 = [movreg(R0,R),idivreg(R1,R0),mulreg(R1,R0),subreg(R0,R)|Pilr],
    release_if_done0(E1,R0,HoldR,Tmid1,Tmid2),
    release_if_done0(E2,R1,HoldR,Tmid2,Tout).
eval_binop(Op,_,_,_,Pil,Pil,_,Tin,Tin) :-
    error(('Unknown function: ', Op, '/2 in is/2')).

eval_numgetinst(VPrag,R,Pil,Pilr,Tin,Tout) :-
	vprag_type(VPrag, t),
	!,
	vprag_loc(VPrag, Loc),
	vprag_occ(VPrag, Occ),
	( ( Occ=l; Occ=v), R = Loc, Pil = Pilr, Tin = Tout
	; getreg(Tin,R), hold(R,Tin,Tout), Pil = [movreg(Loc,R)|Pilr]
	).
eval_numgetinst(Prag,R,Pil,Pilr,Tin,Tout) :-
	getreg(Tin,R), hold(R,Tin,Tout),
	putvarinst(Prag,R,Pil,Pilr,Tout).

force_tp_eval_loaded.

