/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  useinfer.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu, David S. Warren
  Modified by		:  Kostis F. Sagonas, David S. Warren
  Last modification	:  May 21, 1993
========================================================================*/

/************************************************************************/
/* NOTE: This file was seriously changed by Kostis and David during the	*/
/*	 summer of 1992 to speed up the XSB compiler.  These changes	*/
/*	 were mostly focused on predicate infer_usage_of_symbol/4 and	*/
/*	 the predicates that it calls.  Predicate add_if_necessary/3	*/
/*	 was added by David during that period.  those changes made the	*/
/*	 algorithm that infer_usage_of_symbol/4 implements linear on	*/
/*	 the symbols in the symbol table.				*/
/*	    On May of 1993, Kostis made some more improvements on the	*/
/*	 speed of inferencing and took out the calls to the inference	*/
/*	 of the exported types stuff.  An earlier version of XSB (say	*/
/*	 XSB v1.0 Beta or even PSB) should be consulted if it is ever	*/
/*	 decided to put the typing inference stuff back to XSB.		*/
/************************************************************************/

/************************************************************************/
/* useinfer(#Module, -IsModule, -ExportList)				*/
/*	Performs inference of exported predicates/functors/constants,	*/
/*	and inference of unused or undefined predicates.  In doing so,	*/
/*	fills in the following entries for each symbol:			*/
/*		@ one of im(M), ex, local, global (internal is not	*/
/*		  used until preprocessing)				*/
/*		@ one of used, unused					*/
/*		@ 'pred' for predicates used in the body		*/
/*		@ 'undef' if a 'pred' is used but not defined in, or	*/
/*		  imported from another module.				*/
/*	Some warning messages will be printed when doing the inference.	*/
/*	E.g., if no export names are declared, a warning message is	*/
/*	      given, and all symbols are assumed to be exported.	*/
/************************************************************************/

useinfer(Module, IsModule, ExportList) :-
	Module = module(SymTab,_TyList,_DcList,_FnList,_ClList,_QrList),
	extract_symlist(ex, ExList, SymTab),
	sym_insert('_$main', 0, [], SymTab, SymMain),
	( ExList = [] -> 
		IsModule = 0,		% No export list, is a global module
		sym_propin(global, SymMain), 
		infer_export_fill(SymTab, global)
	; IsModule = 1,
	  ( sym_prop(im(_Mod), SymMain) -> true ; sym_propin(local, SymMain) ),
	  infer_usage([SymMain|ExList], Module),
	  infer_export_fill(SymTab, local),
	  infer_fillunused(SymTab)
	),
	get_sym_names(ExList, ExportList).


/*======================================================================*/
/* infer_export_fill(#SymTab, ex/local/global)				*/
/* 	Mark all symbols that do not have an environment field yet	*/
/*	by ex/local/global.						*/
/*======================================================================*/

infer_export_fill(RestTab, Domain) :-
	sym_scan(Sym, RestTab, RestTab2),
	sym_env(Sym, Dom),
	( nonvar(Dom) -> true
	; sym_prop(pred, Sym) -> Dom = Domain
	; Dom = global	% functors are assumed global; temporarily
	),
	infer_export_fill(RestTab2, Domain).
infer_export_fill(_RestTab, _Domain).


/*======================================================================*/
/* infer_usage(+SymList, #Module)					*/
/*	Infers the usage of symbols appearing in a real module.		*/
/*	SymList is a list of symbols to be checked, and the result is	*/
/*	stored in the symbol table.  Information is inserted into the	*/
/*	symbol table when a symbol is used, or when an undefined	*/
/*	predicate is encountered.					*/
/*======================================================================*/

infer_usage([], _Module).
infer_usage([Sym|Rest], Module) :-
	infer_usage_of_symbol(Sym, Rest, NList, Module),
	infer_usage(NList, Module).

%% infer_usage_of_symbol(+P, +A, +Rest, -NewRest, #Module)
%%	Mark P/A "used", and look for possible symbols used by P/A, e.g
%%	predicates used in the clauses of P/A when it is a predicate.
%%	"Rest" is the symbol list remained to be checked. Symbols used
%%	by P/A is added to the list and is returned in "NewRest".
%%	There are the following cases:
%%	(1) P/A is already "used", this means that it is already checked
%%		Nothing needs to do; NewRest is set to Rest.
%%	SymList is a list of symbols to be checked,

infer_usage_of_symbol(Sym, InList, OutList, Module) :-
	Module = module(SymTab,_TyList,_DcList,_FnList,ClList,QrList),
	( sym_prop(used, Sym) -> InList = OutList	% already checked
	; sym_name(Sym, '_$main', 0) -> 
		( QrList=[] -> InList=OutList
		; sym_propin(used, Sym),
		  member(pred(Sym,Cls,_), QrList),
		  check_pred(Cls, SymTab, InList, OutList)
		)
	; sym_propin(used, Sym),
	  ( sym_prop(pred, Sym) ->
	    	( member(pred(Sym,Cls,_), ClList) ->
			check_pred(Cls, SymTab, InList, OutList)
		; sym_prop(im(Mod), Sym) -> 
			sym_insert(Mod, 0, [], SymTab, ModSym), 
			OutList=[ModSym|InList]
		; \+ option(sysmod), sym_name(Sym, P, A),
		  standard_symbol(P, A, Props) -> 
			member(im(Mod), Props),
			sym_insert(Mod, 0, [module,used], SymTab, ModSym),
			sym_propin(Props, Sym),
			OutList=InList
		; sym_name(Sym, P, A), warning(('Undefined predicate ', P/A)), 
		  sym_propin(undef, Sym), OutList=InList
		)
/*	  ; sym_prop(fun, Sym) -> ???  This was commented by Jiyang */
	  ; sym_prop(im(Mod), Sym) ->	% Added by Kostis
		% takes care of the case when Sym is imported from Mod but is
		% not a predicate symbol. If Sym is used, then Mod is used.
		sym_insert(Mod, 0, [], SymTab, ModSym), OutList=[ModSym|InList]
	  ; InList = OutList
		/* include the case of a module, a constant, a functor, ... */
	  )
	).

check_pred([], _SymTab, InList, InList) :- !.
check_pred([Cl|Cls], SymTab, InList, OutList) :-
	check_clause(Cl, SymTab, InList, MedList),
	check_pred(Cls, SymTab, MedList, OutList).

check_clause(clause(Args,Body, _), SymTab, InList, OutList) :-
	check_args(Args, SymTab, InList, MedList),
	check_body(Body, SymTab, MedList, OutList).

check_body(varocc(_), _SymTab, InList, OutList) :- !,
	OutList = InList.
check_body(and(L1,L2), SymTab, InList, OutList) :- !,
	check_body(L1, SymTab, InList, MedList),
	check_body(L2, SymTab, MedList, OutList).
check_body(or(L1,L2), SymTab, InList, OutList) :- !,
	check_body(L1, SymTab, InList, MedList),
	check_body(L2, SymTab, MedList, OutList).
check_body(if(L1,L2), SymTab, InList, OutList) :- !,
	check_body(L1, SymTab, InList, MedList),
	check_body(L2, SymTab, MedList, OutList).
check_body(not(L1), SymTab, InList, OutList) :- !,
	check_body(L1, SymTab, InList, OutList).
check_body(inlinegoal(_,_,Args), SymTab, InList, OutList) :-
	check_args(Args, SymTab, InList, OutList).
check_body(goal(Sym,Args), SymTab, InList, OutList) :-
	sym_propin(pred, Sym),
	add_if_necessary(Sym, InList, MedList),
	check_args(Args, SymTab, MedList, OutList).
check_body(fungoal(Sym,_Res,Args), SymTab, InList, OutList) :-
	add_if_necessary(Sym, InList, MedList),
	check_args(Args, SymTab, MedList, OutList),
	!.
check_body(funbody(_R,L), SymTab, InList, OutList) :- !,
	check_body(L, SymTab, InList, OutList).


check_args([], _SymTab, InList, InList).
check_args([Arg|Args], SymTab, InList, OutList) :-
	check_arg(Arg, SymTab, InList, MedList),
	check_args(Args, SymTab, MedList, OutList).

check_arg(constant(Sym), _SymTab, InList, [Sym|InList]).
check_arg(integer(_), _SymTab, InList, InList).
check_arg(real(_), _SymTab, InList, InList).
check_arg(varocc(_), _SymTab, InList, InList).
check_arg(structure(Sym,Args), SymTab, InList, OutList) :-
	add_if_necessary(Sym, InList, MedList),
	check_args(Args, SymTab, MedList, OutList).


/*======================================================================*/
/* infer_fillunused(#SymTab)						*/
/*	Mark all symbols that are not "used" as "unused".		*/
/*======================================================================*/

infer_fillunused(SymTab) :-
	sym_scan(Sym, SymTab, RestTab),
	( sym_prop(used, Sym) -> true 
	; sym_propin(unused, Sym), sym_name(Sym, N, A),
	  warning(('Unused symbol: ', N/A))
	),
	infer_fillunused(RestTab).
infer_fillunused(_).


/*======================================================================*/
/* add_if_necessary(+Sym, +InList, -OutList)				*/
/*	Try to add symbol Sym in list InList giving OutList only when	*/
/*	it is necessary.						*/
/*======================================================================*/

add_if_necessary(Sym, InList, OutList) :-
	( sym_prop(used,Sym) -> OutList = InList
	; ( sym_prop(pred,Sym) -> OutList = [Sym|InList]
	  ; ( sym_prop(im(_Mod),Sym) -> OutList = [Sym|InList]
	    ; ( sym_prop((fun),Sym) -> OutList = [Sym|InList]
	      ;	OutList = InList, sym_propin(used,Sym)
	      )
	    )
	  )
	).

/*======================================================================*/
/* get_sym_names(+SymList, -PredList)					*/
/*	From a SymList we get a PredList.				*/
/*======================================================================*/

get_sym_names([], []).
get_sym_names([Sym|ExportList0], [N/A|ExportList]) :-
        sym_name(Sym, N, A),
        get_sym_names(ExportList0, ExportList).

