/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  eval.P
  Author(s)		:  David S. Warren, Kostis F. Sagonas
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  September 14, 1992
========================================================================*/


:- export eval/2.

:- import term_type/2, term_psc/2, psc_type/2, 
	  psc_prop/2, code_call/3, psc_name/2
     from machine.
:- import functor/3, arg/3, abort/1 from standard.
:- import load/1 from loader.


%-------------------------------------------------------------------------
% NOTE:	Predicate eval/2 is the one who is responsible for error checking
%	of arithmetic comparisons and evaluations. According to the Prolog
%	standard, arithmetic comparisons and evaluations can cause
%	instantiation errors ONLY. Everything else should simply fail as
%	far as evaluation is concerned.
%-------------------------------------------------------------------------
% NOTE: According to the Prolog standard, predicate eval/2 should perform 
%	type conversion between integers and reals when necessary. 
%	Our eval/2, does NOT. (mainly because at the moment when this is
%	written it is not clear whether we will support reals in the first
%	XSB version).
%-------------------------------------------------------------------------

eval(Arg, Res) :- 
	term_type(Arg, T), 
	( T =:= 0 -> abort('Unbound variable in arithmetic expression')
	; T =:= 2 -> Res=Arg
	; T =:= 6 -> Res=Arg
	; eval0(Arg, Res) -> true	/* system defined function */
	; T =:= 1 -> eval1(Arg, Res)	/* user defined function */
	).

eval0('+'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 + Res2.
eval0('-'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 - Res2.
eval0('*'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 * Res2.
eval0('/'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 / Res2.
eval0('//'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 // Res2.
eval0(mod(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 mod Res2.
eval0('-'(X), Res) :- eval(X, Res1), Res is -(Res1).
eval0('/\'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 /\ Res2.
eval0('\/'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 \/ Res2.
eval0('<<'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 << Res2.
eval0('>>'(X,Y), Res) :- eval(X, Res1), eval(Y, Res2), Res is Res1 >> Res2.
eval0('\'(X), Res) :- eval(X, Res1), Res is '\'(Res1).
eval0(float(X), Res) :- eval(X, Res1), Res is float(Res1).
eval0(floor(X), Res) :- eval(X, Res1), Res is floor(Res1).
eval0(sin(X), Res) :- eval(X, Res1), Res is sin(Res1).
eval0(cos(X), Res) :- eval(X, Res1), Res is cos(Res1).
eval0(tan(X), Res) :- eval(X, Res1), Res is tan(Res1).
eval0(asin(X), Res) :- eval(X, Res1), Res is asin(Res1).
eval0(acos(X), Res) :- eval(X, Res1), Res is acos(Res1).
eval0(atan(X), Res) :- eval(X, Res1), Res is atan(Res1).
eval0(exp(X), Res) :- eval(X, Res1), Res is exp(Res1).
eval0(log(X), Res) :- eval(X, Res1), Res is log(Res1).
eval0(log10(X), Res) :- eval(X, Res1), Res is log10(Res1).
eval0(sqrt(X), Res) :- eval(X, Res1), Res is sqrt(Res1).

eval1(Term, Res) :-
	term_psc(Term, Psc),
	psc_type(Psc, T0),
	( T0 =:= 15 ->		% unloaded function
	    psc_prop(Psc, ModPSC),
            psc_name(ModPSC, ModName),
            psc_type(ModPSC, ModType),
            ( ModType =:= 4
            ; ModType =\= 4,              % module is not loaded yet
              load(ModName)               % can fail here
            )
	; true
	),
	psc_type(Psc, T1),
	( T1 =\= 14, eval_regularterm(Term, Res)	 % Regular term
	; T1 =:= 14,
          psc_prop(Psc, Code),
	  eval_func(Code, Term, T1),
	  '_$getreturn'(Res)
	).

/* need this because code_call must be the last goal of a clause */
eval_func(Code, Term, T1) :- code_call(Code, Term, T1).

eval_regularterm(Term, Res) :-
	functor(Term, F, A),
	functor(Res, F, A),
	eval_rt_args(0, A, Term, Res).

eval_rt_args(A, A, _, _).
eval_rt_args(I, A, Term, Res) :- I < A,
	I1 is I + 1,
	arg(I1, Term, X),
	arg(I1, Res, Y),
	eval(X, Y),
	eval_rt_args(I1, A, Term, Res).
