/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  setof.P
  Author(s)		:  Apparently nobody wants to admit writing this
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  September 8, 1992
========================================================================*/


/* setof : 'setof/3', 'bagof/3', 'findall/3' and sorting. */
/* This was taken from the C-Prolog system, and modified to use $findall,
   instead of `record' */


:- export setof/3, bagof/3, findall/3, sort/2, keysort/2, '^'/2.


:- import length/2 from basics.
:- import copyterm/6 from copyterm.
:- import functor/3, (=..)/2, arg/3, compare/3, '\=='/2, '=='/2,
	  call/1, ttywritenl/1, call_expose/1, is_list/1, abort/1
     from standard.
:- import buff_word/3, buff_set_word/3, buff_alloc/3, stat_flag/2,
	  buff_dealloc/4, buff_cell/3 %, lock_init/1, lock_on/1, lock_off/1
     from machine.


setof(Template, Generator, Set) :-
	bagof(Template, Generator, Bag),
	sort(Bag, Set0),
	Set=Set0.


bagof(Template, Generator, Bag) :-
	excess_vars(Generator, Template, [], VarList), 
	nonempty(VarList), !,
	Key =.. ['$'|VarList],
	findall(Key-Template, Generator, Bags0),
	keysort(Bags0, Bags),
	pick(Bags, Key, Bag).
bagof(Template, Generator, Bag) :-
	findall(Template, Generator, Bag),
	nonempty(Bag).	% If Bag=[] then bagof/3 and setof/3 should fail.


nonempty([_|_]).


pick(Bags,Key,Bag) :-
   nonempty(Bags),
   parade(Bags,Key1,Bag1,Bags1),
   decide(Key1,Bag1,Bags1,Key,Bag).

parade([Item|L1],K,[X|B],L) :- item(Item,K,X), !,
   parade(L1,K,B,L).
parade(L,_,[],L).

item(K-X,K,X).

decide(Key,Bag,Bags,Key,Bag) :- (Bags=[], ! ; true).
decide(_,_,Bags,Key,Bag) :- pick(Bags,Key,Bag).

excess_vars(T,X,L0,L) :- var(T), !,
   ( no_occurrence(T,X), !, introduce(T,L0,L)
   ; L = L0 ).
excess_vars(X^P,Y,L0,L) :- !, excess_vars(P,(X,Y),L0,L).
excess_vars(setof(X,P,S),Y,L0,L) :- !, excess_vars((P,S),(X,Y),L0,L).
excess_vars(bagof(X,P,S),Y,L0,L) :- !, excess_vars((P,S),(X,Y),L0,L).
excess_vars(T,X,L0,L) :- functor(T,_,N), rem_excess_vars(N,T,X,L0,L).

rem_excess_vars(0,_,_,L,L) :- !.
rem_excess_vars(N,T,X,L0,L) :-
   arg(N,T,T1),
   excess_vars(T1,X,L0,L1),
   N1 is N-1,
   rem_excess_vars(N1,T,X,L1,L).

introduce(X,L,L) :- included(X,L), !.
introduce(X,L,[X|L]).

included(X,L) :- doesnt_include(L,X), !, fail.
included(_,_).

doesnt_include([],_).
doesnt_include([Y|L],X) :- Y \== X, doesnt_include(L,X).

no_occurrence(X,Term) :- contains(Term,X), !, fail.
no_occurrence(_,_).

contains(T,X) :- var(T), !, T == X.
contains(T,X) :- functor(T,_,N), upto(N,I), arg(I,T,T1), contains(T1,X).

upto(N,N) :- N > 0.
upto(N,I) :- N > 0, N1 is N-1, upto(N1,I).

/* -------------------------------------------------------------------- */
/* Sorting a list by bisecting and merging. 				*/
/* -------------------------------------------------------------------- */

sort(L, R) :- 
	( is_list(L) -> length(L, N), sort(N, L, _, R1), R = R1
	; abort('Uninstantiated argument 1 of sort/2, cannot sort!')
	).

sort(2,[X1|L1],L,R) :- !, comprises(L1,X2,L),
    compare(Delta,X1,X2),
  (Delta = (<) , !, R = [X1,X2]
   ; Delta = (>) , !, R = [X2,X1]
   ; R = [X2]
  ).
sort(1,[X|L],L,[X]) :- !.
sort(0,L,L,[]) :- !.
sort(N,L1,L3,R) :-
   N1 is N//2, N2 is N-N1,
   sort(N1,L1,L2,R1),
   sort(N2,L2,L3,R2),
   merge(R1,R2,R).

merge([],R,R) :- !.
merge(R,[],R) :- !.
merge(R1,R2,[X|R]) :-
   comprises(R1,X1,R1a), comprises(R2,X2,R2a),
   compare(Delta,X1,X2),
  (Delta = (<) , !, X = X1, merge(R1a,R2,R)
   ; Delta = (>) , !, X = X2, merge(R1,R2a,R)
   ; X = X1, merge(R1a,R2a,R)
  ).

comprises([X|L],X,L).

/* -------------------------------------------------------------------- */
/* Sorting on keys by bisecting and merging.				*/
/* -------------------------------------------------------------------- */

keysort(L,R) :- length(L,N), keysort(N,L,_,R1), R=R1.

keysort(2,[X1|L1],L,R) :- !,
   comprises(L1,X2,L),
   compare_keys(Delta,X1,X2),
  (Delta = (>) , !, R = [X2,X1] ; R = [X1,X2] ).
keysort(1,[X|L],L,[X]) :- !.
keysort(0,L,L,[]) :- !.
keysort(N,L1,L3,R) :-
   N1 is N//2, N2 is N-N1,
   keysort(N1,L1,L2,R1),
   keysort(N2,L2,L3,R2),
   keymerge(R1,R2,R).

keymerge([],R,R) :- !.
keymerge(R,[],R) :- !.
keymerge(R1,R2,[X|R]) :-
   comprises(R1,X1,R1a), comprises(R2,X2,R2a),
   compare_keys(Delta,X1,X2),
  (Delta = (>) , !, X = X2, keymerge(R1,R2a,R)
   ; X = X1, keymerge(R1a,R2,R)
  ).

compare_keys(Delta, K1-_X1, K2-_X2) :- compare(Delta, K1, K2).

/*======================================================================*/

_X^P :- call(P).

/*======================================================================*/
/* findall(+Term, +Call, -Result)					*/
/*	The first word in buffer is the size of the buffer; not used	*/
/*	The second word is the pointer to the argument position where	*/
/*		the current answer should be be put.			*/
/*		This location is initialized to [] before writing.	*/
/*	The third word is a pointer to the next free location in buffer	*/
/*	The rest of buffer contains the answer (a list)			*/
/*======================================================================*/

findall(T, Call, Result) :-
	buff_alloc(65536, Buff, 0),
	findall_1(T, Call, Result, Buff, 65536).

findall_1(T, Call, _Result, Buff, BuffLen) :-
	wordsize(Wordsize),
	copyterm([], Buff, 8, _, _, BuffLen), /* init result list to empty */
	buff_set_word(Buff, 0, 8), /* init where to put next answer */
	FirstFree is 8 + Wordsize,
	buff_set_word(Buff, 4, FirstFree), /* init first free place */
    Place=Place, Start=Start,End=End,
	%%lock_init(Lock),
	call_expose(Call),
	%%lock_on(Lock),
	buff_word(Buff, 0, Place), /* get where to put answer */
	buff_word(Buff, 4, Start), /* get first free place */
	copyterm([T], Buff, Place, Start, End, BuffLen),
	Tailloc is Start+Wordsize,
	buff_set_word(Buff, 0, Tailloc), /* where to put next answer */
				/* point to the tail [] */
	buff_set_word(Buff, 4, End), /* next first free place */
	%%lock_off(Lock),
	End > BuffLen, ttywritenl('Findall buffer overflow'),
	fail.
findall_1(_, _, Result, Buff, BuffLen) :-
	buff_word(Buff, 4, Length),	/* Length =\= 12 fail if [] */
	buff_dealloc(Buff, BuffLen, Length, 0),
	buff_cell(Buff, 8, Result0),
	Result = Result0.

/*======================================================================*/
/* function that determines the size of a word.                         */
/*      will be implemented as function. But this librury is made       */
/*      by V 3.3, which does not support functions.                     */
/*======================================================================*/

wordsize(8) :- stat_flag(18, Word), Word = 3, !.
wordsize(4).
