/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  dcg_qsc.P
  Author(s)		:  Kostis F. Sagonas
  Last modification	:  September 14, 1992
  File Description	:  A straightforward translator for 
			   Definite Clause Grammars as is being done in
			   Quintus Prolog, SISCtus Prolog and C-Prolog.
========================================================================*/


/*----------------------------------------------------------------------*/
/*  dcg(+DCG_Rule, -Prolog_Clause)					*/
/*  Prolog_Clause is the corresponding expanded version of DCG_Rule.	*/
/*  Predicate dcg/2 is called by expand_term/2 when there is not a	*/
/*  user-defined predicate term_expansion/2. Predicate dcg/2 performs	*/
/*  the error checking for expand_term/2.				*/
/*  NOTE: This DCG expansion is similar to the one found in Quintus	*/
/*	  Prolog, SICStus Prolog and C-Prolog. Its difference from	*/
/*	  the one found in file dcg.P is that the expansion is not	*/
/*	  guaranteed to be steadfast in all cases when the clause to	*/
/*	  be expanded contains a '!'. To find out more about this	*/
/*	  difference read the User's manual.				*/
/*----------------------------------------------------------------------*/

dcg((LHS --> RHS), (OutHead :- OutBody)) :-
	( functor(LHS, ',', 2) -> 
		arg(1, LHS, Non_Terminal),
		( var(Non_Terminal) ->
			abort('Uninstantiated left-hand side of DCG rule')
		; arg(2, LHS, PushBack_List),
		  ( is_list(PushBack_List) -> true
		  ; abort('Push-Back list in DCG rule is not a proper list')
		  )
		),
		New_LHS = LHS
	; ( var(LHS) -> abort('Uninstantiated left-hand side of DCG rule') ; true ),
	  New_LHS = ','(LHS, [])
	),
	dcg_with_pushback_list(New_LHS, RHS, OutHead, OutBody).

dcg_with_pushback_list(','(Non_Terminal, []), RHS, OutHead, OutBody) :-
	!,
	expand_dcg_non_ternimal(Non_Terminal, In, Out, OutHead),
	expand_dcg_body(RHS, In, Out, OutBody).
dcg_with_pushback_list(','(Non_Terminal, PushBack_List), 
                       RHS, 
                       OutHead, 
                       ','(Normal_Body, PushBack_Body)) :-
	expand_dcg_non_ternimal(Non_Terminal, In, Out, OutHead),
	expand_dcg_body(RHS, In, Middle, Normal_Body),
	dcg_list_of_terminals(PushBack_List, Out, Middle, PushBack_Body).

expand_dcg_non_ternimal(Non_Terminal, In, Out, NewTerm) :-
	functor(Non_Terminal, P, Arity),
	NewArity is Arity + 2,
	functor(NewTerm, P, NewArity),
	N1 is Arity + 1,
	arg(N1, NewTerm, In),
	arg(NewArity, NewTerm, Out),
	dcg_fillin_arguments(Arity, Non_Terminal, NewTerm).

dcg_fillin_arguments(N, DCG_Term, Prolog_Term) :-
	( N =:= 0 -> true 
	; arg(N, DCG_Term, Arg), arg(N, Prolog_Term, Arg),
	  NewN is N - 1, dcg_fillin_arguments(NewN, DCG_Term, Prolog_Term)
	).

expand_dcg_body(RHS, In, Out, Goal) :-
	( (RHS == '!' ; RHS == '{}'('!')) -> 
		Goal = ('!', In = Out)	/* Make the expansion steadfast. */
	; dcg_body(RHS, In, Out, Goal)
	).

dcg_body(RHS, In, Out, Goal) :-
	( var(RHS) -> dcg_body_var(RHS, In, Out, Goal) 
	; dcg_body_nonvar(RHS, In, Out, Goal)
	).

% NOTE:	What the following fact should return is really:
%		phrase(user:Phrase, In, Out)
%	but for the moment I dunno the way to do it.
dcg_body_var(Phrase, In, Out, phrase(Phrase, In, Out)).

dcg_body_nonvar(','(Term1,Term2), In, Out, ','(Goal1,Goal2)) :-
	!,
	dcg_body(Term1, In, Middle, Goal1),
	dcg_body(Term2, Middle, Out, Goal2).
dcg_body_nonvar(';'(Term1,Term2), In, Out, ';'(Goal1,Goal2)) :-
	!,
	dcg_body(Term1, In, Out, Goal1),
	dcg_body(Term2, In, Out, Goal2).
dcg_body_nonvar('->'(Term1,Term2), In, Out, '->'(Goal1,Goal2)) :-
	!,
	dcg_body(Term1, In, Middle, Goal1),
	dcg_body(Term2, Middle, Out, Goal2).
dcg_body_nonvar(not(Term), In, Out, not(Goal)) :-
	!,
	warning('Use of not/1 in DCG rule may not be sound!'),
	dcg_body(Term, In, Out, Goal).
dcg_body_nonvar('\+'(Term), In, Out, '\+'(Goal)) :-
	!,
	warning('Use of (\+)/1 in DCG rule may not be sound!'),
	dcg_body(Term, In, Out, Goal).
dcg_body_nonvar(fail_if(Term), In, Out, fail_if(Goal)) :-
	!,
	warning('Use of fail_if/1 in DCG rule may not be sound!'),
	dcg_body(Term, In, Out, Goal).
dcg_body_nonvar('{}'(G), In, Out, Goal) :- 
	!,
	( G == '!' -> Goal = G, Out = In	/* A {!} is the same as a !. */
		    ; Goal = (G, In = Out) ).
dcg_body_nonvar('!', In, In, '!') :- 
	!.
dcg_body_nonvar([], In, Out, (In = Out)) :-
	!.
dcg_body_nonvar(L, In, Out, Goal) :-
	( (L = [_|_] ->
		( is_list(L) -> 		/* proper list of terminals */
			dcg_list_of_terminals(L, In, Out, Goal)
		; abort('A list of terminals in DCG rule is not a proper list')
		)
	  )
	; expand_dcg_non_ternimal(L, In, Out, Goal)	/* nonterminal */
	).

dcg_list_of_terminals([H], In, Out, 'C'(In,H,Out)).
dcg_list_of_terminals([H|L], In, Out, ','('C'(In,H,Mid), Goal)) :-
     dcg_list_of_terminals(L, Mid, Out, Goal).

/*----------------------------------------------------------------------*/
/*  'C'(S1, T, S2)	: S1 is connected to S2 by the terminal T.	*/
/*----------------------------------------------------------------------*/

'C'([Token|Tokens], Token, Tokens).

/*----------------------------------------------------------------------*/
/*  phrase(P, L)	: list L can be parsed as a phrase of type P.	*/
/*  phrase(P, L, R) 	: list R is what remains of list L after phrase */
/*			  P has been found.				*/
/*----------------------------------------------------------------------*/

phrase(Phrase_Type, List) :-
	( nonvar(Phrase_Type) -> phrase(Phrase_Type, List, []) 
	; abort('Uninstantiated argument 1 of phrase/2')
	).

phrase(Phrase_Type, List, Rest) :-
	( nonvar(Phrase_Type) ->
		dcg_body_nonvar(Phrase_Type, List, Rest, Goal),
		call(Goal)
	; abort('Uninstantiated argument 1 of phrase/3')
	).

/* -------------------- end of file dcg_qsc.P ------------------------- */

